# (c) Copyright 2018 Palantir Technologies Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from requests.adapters import HTTPAdapter, Response, CaseInsensitiveDict
from typing import TypeVar, Type, List, Optional, Dict, Any, Union
from requests.exceptions import HTTPError
from requests.packages.urllib3.poolmanager import PoolManager
from requests.packages.urllib3.util.ssl_ import create_urllib3_context
from requests.packages.urllib3.util import Retry
from .configuration import ServiceConfiguration

import binascii
import os
import random
import requests


T = TypeVar("T")
# https://testssl.sh/openssl-rfc.mappping.html
# https://wiki.mozilla.org/Security/Server_Side_TLS#Intermediate_compatibility_.28recommended.29
CIPHERS = (
    # tls 1.3 ciphers
    # n.b. these may not actually be honoured given the comment in:
    # https://docs.python.org/3/library/ssl.html#ssl.SSLContext.set_ciphers
    "TLS_AES_128_GCM_SHA256:"
    "TLS_AES_256_GCM_SHA384:"
    "TLS_CHACHA20_POLY1305_SHA256:"
    # non-tls 1.3 ciphers
    "ECDHE-ECDSA-AES128-GCM-SHA256:"
    "ECDHE-RSA-AES128-GCM-SHA256:"
    "ECDHE-ECDSA-AES256-GCM-SHA384:"
    "ECDHE-RSA-AES256-GCM-SHA384:"
    "ECDHE-ECDSA-CHACHA20-POLY1305:"
    "ECDHE-RSA-CHACHA20-POLY1305:"
    "DHE-RSA-AES128-GCM-SHA256:"
    "DHE-RSA-AES256-GCM-SHA384"
)


TRACE_ID_HEADER: str = "X-B3-TraceId"
TRACE_ID_RANDOM_BYTES = 8


def fresh_trace_id() -> bytes:
    # returns a bytestring which is a valid zipkin trace id
    return binascii.hexlify(os.urandom(TRACE_ID_RANDOM_BYTES))


class Service(object):
    _requests_session: requests.Session
    _uris: List[str]
    _connect_timeout: float
    _read_timeout: float
    _verify: str

    def __init__(
        self,
        requests_session: requests.Session,
        uris: List[str],
        _connect_timeout: float,
        _read_timeout: float,
        _verify: str,
    ) -> None:
        self._requests_session = requests_session
        self._uris = uris
        self._connect_timeout = _connect_timeout
        self._read_timeout = _read_timeout
        self._verify = _verify

    @property
    def _uri(self) -> str:
        """returns a random uri"""
        return random.SystemRandom().choice(self._uris)

    def _request(self, *args, **kwargs) -> Response:
        """Make requests using configured :class:`requests.Session`.
        Any error details will be extracted to an :class:`HTTPError`
        which will contain relevant error details when printed."""
        self._amend_request_kwargs(kwargs)
        _response = self._requests_session.request(*args, **kwargs)
        try:
            _response.raise_for_status()
        except HTTPError as e:
            if e.response is not None:
                raise ConjureHTTPError(e) from e
            raise e
        return _response

    def __repr__(self) -> str:
        return "{}(requests_session={}, uris={})".format(
            self.__class__.__name__, "requests.Session(...)", repr(self._uris)
        )

    def _amend_request_kwargs(self, kwargs) -> None:
        for param_kind in ["headers", "params"]:
            if param_kind in kwargs:
                kwargs[param_kind] = _clean_params(kwargs[param_kind])
        kwargs["timeout"] = (self._connect_timeout, self._read_timeout)
        kwargs["verify"] = self._verify
        _add_trace_id(kwargs)


def _clean_params(params: Dict[str, Any]) -> Dict[str, Any]:
    cleaned_params = {}
    for key, value in params.items():
        if value is None:
            continue
        cleaned_params[key] = _clean_param_value(value)
    return cleaned_params


def _clean_param_value(value: Any) -> Any:
    if isinstance(value, list):
        return [_clean_param_value(element) for element in value]
    if isinstance(value, bool):
        return str(value).lower()
    return str(value)


def _add_trace_id(kwargs: Dict[str, Any]) -> None:
    # Adds the trace ID to the arguments
    if "headers" not in kwargs:
        kwargs["headers"] = {}
    kwargs["headers"][TRACE_ID_HEADER] = fresh_trace_id()


class RetryWithJitter(Retry):
    """
    Extends the standard urllib Retry in order to match conjure behaviour.
    Retry times contain a uniform jitter. Additionally, all http methods are
    considered valid to retry.
    """

    @classmethod
    def _is_method_retryable(cls, method):
        return True

    def get_backoff_time(self):
        jitter = random.random()
        return jitter * super().get_backoff_time()


class RequestsClient(object):
    @classmethod
    def create(
        cls,
        service_class: Type[T],
        user_agent: str,
        service_config: ServiceConfiguration,
    ) -> T:
        # setup retry to match java remoting
        # https://github.com/palantir/http-remoting/tree/3.12.0#quality-of-service-retry-failover-throttling
        retry = RetryWithJitter(
            total=service_config.max_num_retries,
            read=0,  # do not retry read errors
            status_forcelist=[308, 429, 503],
            backoff_factor=float(service_config.backoff_slot_size) / 1000,
        )
        transport_adapter = TransportAdapter(max_retries=retry)
        # create a session, for shared connection polling, user agent, etc
        session = requests.Session()
        session.headers = CaseInsensitiveDict({"User-Agent": user_agent})
        if service_config.security is not None:
            verify = service_config.security.trust_store_path
        else:
            verify = None
        for uri in service_config.uris:
            session.mount(uri, transport_adapter)
        return service_class(  # type: ignore
            session,
            service_config.uris,
            service_config.connect_timeout,
            service_config.read_timeout,
            verify,
        )


class TransportAdapter(HTTPAdapter):
    """Transport adapter that allows customising ssl things"""

    def init_poolmanager(
        self, connections, maxsize, block=False, **pool_kwargs
    ):
        self._pool_connections = connections
        self._pool_maxsize = maxsize
        self._pool_block = block
        ssl_context = create_urllib3_context(ciphers=CIPHERS)
        self.poolmanager = PoolManager(
            num_pools=connections,
            maxsize=maxsize,
            block=block,
            strict=True,
            ssl_context=ssl_context,
            **pool_kwargs
        )


class ConjureHTTPError(HTTPError):
    """An HTTPError from a Conjure Service with ``SerializableError``
    attributes extracted from the response."""

    _cause: Optional[HTTPError]
    _error_code: str
    _error_name: str
    _error_instance_id: str
    _parameters: Dict[str, str]
    _trace_id: str

    def __init__(self, http_error: HTTPError) -> None:
        self._cause = http_error
        self._trace_id = http_error.response.headers.get("X-B3-TraceId")
        try:
            detail = http_error.response.json()
            self._error_code = detail.get("errorCode")
            self._error_name = detail.get("errorName")
            self._error_instance_id = detail.get("errorInstanceId")
            self._parameters = detail.get("parameters", dict())
            message = (
                "{}. ErrorCode: '{}'. ErrorName: '{}'. "
                "ErrorInstanceId: '{}'. TraceId: '{}'. Parameters: {}".format(
                    http_error,
                    self._error_code,
                    self._error_name,
                    self._error_instance_id,
                    self._trace_id,
                    self._parameters,
                )
            )
        except ValueError:
            message = "{}. TraceId: '{}'. Response: '{}'".format(
                http_error, self._trace_id, http_error.response.text
            )
        super(ConjureHTTPError, self).__init__(
            message, request=http_error.request, response=http_error.response
        )

    @property
    def cause(self) -> Optional[HTTPError]:
        """The wrapped ``HTTPError`` that was the direct cause of
        the ``ConjureHTTPError``.
        """
        return self._cause

    @property
    def error_code(self) -> str:
        """A fixed code word identifying the type of error."""
        return self._error_code

    @property
    def error_name(self) -> str:
        """A fixed name identifying the error."""
        return self._error_name

    @property
    def error_instance_id(self) -> str:
        """A unique identifier for this error instance."""
        return self._error_instance_id

    @property
    def parameters(self) -> Dict[str, str]:
        """A set of parameters that further explain the error."""
        return self._parameters

    @property
    def trace_id(self) -> str:
        """The X-B3-TraceId for the request."""
        return self._trace_id



# (c) Copyright 2018 Palantir Technologies Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum
from typing import List, Dict, Type, Any, Union, Optional, TypeVar, Generic

from .case import to_snake_case
from .sanitize import sanitize_identifier


class ConjureType(object):
    pass


DecodableType = Union[
    object,
    int,
    float,
    bool,
    str,
    ConjureType,
    List[Any],
    Dict[Any, Any],
]


T = TypeVar("T", bound=DecodableType)


class OptionalTypeWrapper(ConjureType, Generic[T]):
    pass


class ConjureFieldDefinition:
    identifier: str
    field_type: Type[DecodableType]

    def __init__(
        self, identifier: str, field_type: Type[DecodableType]
    ) -> None:
        self.identifier = identifier
        self.field_type = field_type


class ListType(ConjureType):
    item_type: Type[DecodableType]

    def __init__(self, item_type: Type[DecodableType]) -> None:
        self.item_type = item_type


class DictType(ConjureType):
    key_type: Type[DecodableType]
    value_type: Type[DecodableType]

    def __init__(
        self, key_type: Type[DecodableType], value_type: Type[DecodableType]
    ) -> None:
        self.key_type = key_type
        self.value_type = value_type


class OptionalType(ConjureType):
    item_type: Type[DecodableType]

    def __init__(self, item_type: Type[DecodableType]) -> None:
        self.item_type = item_type


class BinaryType(ConjureType):
    pass


class ConjureEnumType(ConjureType, Enum):

    # override __format__ of ConjureEnumType to prevent issue with default
    # Enum behaviour due to lack of __format__ method on ConjureType
    def __format__(self, format_spec):
        return self.__str__()

    def __repr__(self):
        return "{}.{}".format(self.__class__.__name__, self.value)


class ConjureBeanType(ConjureType):
    @classmethod
    def _fields(cls) -> Dict[str, ConjureFieldDefinition]:
        """_fields is a mapping from constructor argument
        name to the field definition"""
        return {}

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, self.__class__):
            return False

        for attr in self._fields():
            if getattr(self, attr) != getattr(other, attr):
                return False

        return True

    def __ne__(self, other: Any) -> bool:
        return not self == other

    def __repr__(self) -> str:
        fields = [
            "{}={}".format(attr, repr(getattr(self, attr)))
            for attr, field_def in self._fields().items()
        ]
        return "{}({})".format(self.__class__.__name__, ", ".join(fields))


class ConjureUnionType(ConjureType):
    _type: str

    @property
    def type(self) -> str:
        """the member name present in this union"""
        return self._type

    @classmethod
    def _options(cls) -> Dict[str, ConjureFieldDefinition]:
        """_options defines a mapping from each member in the union
        to the field definition for that type"""
        return {}

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, self.__class__):
            return False

        assert isinstance(other, ConjureUnionType)

        pythonic_sanitized_identifier = sanitize_identifier(
            to_snake_case(self.type)
        )

        return other.type == self.type and getattr(
            self, pythonic_sanitized_identifier
        ) == getattr(other, pythonic_sanitized_identifier)

    def __ne__(self, other: Any) -> bool:
        return not self == other

    def __repr__(self) -> str:
        fields = [
            "{}={}".format(attr, repr(getattr(self, attr)))
            for attr, field_def in self._options().items()
            if getattr(self, attr) is not None
        ]
        return "{}({})".format(self.__class__.__name__, ", ".join(fields))



# (c) Copyright 2018 Palantir Technologies Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .._lib import (
    ConjureBeanType,
    ConjureEnumType,
    ConjureUnionType,
    DecodableType,
    BinaryType,
    OptionalTypeWrapper,
    DictType,
    ListType,
    OptionalType,
)
from typing import Optional, Type, Union, get_origin, get_args
from typing import Dict, Any, List
import inspect
import json

NoneType = type(None)


class ConjureDecoder(object):
    """Decodes json into a conjure object"""

    @classmethod
    def decode_conjure_bean_type(cls, obj, conjure_type):
        """Decodes json into a conjure bean type (a plain bean, not enum
        or union).

        Args:
            obj: the json object to decode
            conjure_type: a class object which is the bean type
                we're decoding into
        Returns:
            A instance of a bean of type conjure_type.
        """
        deserialized: Dict[str, Any] = {}
        for (
            python_arg_name,
            field_definition,
        ) in conjure_type._fields().items():
            field_identifier = field_definition.identifier

            if field_identifier not in obj or obj[field_identifier] is None:
                cls.check_null_field(
                    obj, deserialized, python_arg_name, field_definition
                )
            else:
                value = obj[field_identifier]
                field_type = field_definition.field_type
                deserialized[python_arg_name] = cls.do_decode(
                    value, field_type
                )
        return conjure_type(**deserialized)

    @classmethod
    def check_null_field(
        cls, obj, deserialized, python_arg_name, field_definition
    ):
        type_origin = get_origin(field_definition.field_type)
        if isinstance(field_definition.field_type, ListType):
            deserialized[python_arg_name] = []
        elif isinstance(field_definition.field_type, DictType):
            deserialized[python_arg_name] = {}
        elif isinstance(field_definition.field_type, OptionalType):
            deserialized[python_arg_name] = None
        elif type_origin is OptionalTypeWrapper:
            deserialized[python_arg_name] = None
        elif type_origin is list:
            deserialized[python_arg_name] = []
        elif type_origin is dict:
            deserialized[python_arg_name] = {}
        else:
            raise Exception(
                "field {} not found in object {}".format(
                    field_definition.identifier, obj
                )
            )

    @classmethod
    def decode_conjure_union_type(cls, obj, conjure_type):
        """Decodes json into a conjure union type.

        Args:
            obj: the json object to decode
            conjure_type: a class object which is the union type
                we're decoding into
        Returns:
            An instance of type conjure_type.
        """
        type_of_union: str = obj["type"]
        for attr, conjure_field in conjure_type._options().items():
            if conjure_field.identifier == type_of_union:
                attribute = attr
                conjure_field_definition = conjure_field
                break
        else:
            raise ValueError(
                "unknown union type {0} for {1}".format(
                    type_of_union, conjure_type
                )
            )

        deserialized: Dict[str, Any] = {}
        if type_of_union not in obj or obj[type_of_union] is None:
            cls.check_null_field(
                obj, deserialized, attribute, conjure_field_definition
            )
        else:
            value = obj[type_of_union]
            field_type = conjure_field_definition.field_type
            deserialized[attribute] = cls.do_decode(value, field_type)
        return conjure_type(**deserialized)

    @classmethod
    def decode_conjure_enum_type(cls, obj, conjure_type):
        """Decodes json into a conjure enum type.

        Args:
            obj: the json object to decode
            conjure_type: a class object which is the enum type
                we're decoding into.
        Returns:
            An instance of enum of type conjure_type.
        """
        if not (isinstance(obj, str) or str(type(obj)) == "<type 'unicode'>"):
            raise Exception(
                "Expected to find str type but found {} instead".format(
                    type(obj)
                )
            )

        if obj in conjure_type.__members__:
            return conjure_type[obj]

        else:
            return conjure_type["UNKNOWN"]

    @classmethod
    def decode_dict(
        cls,
        obj: Dict[Any, Any],
        key_type: Type[DecodableType],
        item_type: Type[DecodableType],
    ) -> Dict[Any, Any]:
        """Decodes json into a dictionary, handling conversion of the
        keys/values (the keys/values may themselves require conversion).

        Args:
            obj: the json object to decode
            key_type: a class object which is the conjure type
                of the keys in this dict
            item_type: a class object which is the conjure type
                of the values in this dict
        Returns:
            A python dictionary, where the keys are instances of type key_type
            and the values are of type value_type.
        """
        if not isinstance(obj, dict):
            raise Exception("expected a python dict")
        if (
            key_type is str
            or isinstance(key_type, BinaryType)
            or key_type is BinaryType
            or (
                inspect.isclass(key_type)
                and issubclass(key_type, ConjureEnumType)
            )
        ):
            return dict(
                (
                    (
                        cls.do_decode(x[0], key_type),
                        cls.do_decode(x[1], item_type),
                    )
                    for x in obj.items()
                )
            )

        return dict(
            (
                (
                    cls.do_decode(json.loads(x[0]), key_type),
                    cls.do_decode(x[1], item_type),
                )
                for x in obj.items()
            )
        )

    @classmethod
    def decode_list(
        cls, obj: List[Any], element_type: Type[DecodableType]
    ) -> List[Any]:
        """Decodes json into a list, handling conversion of the elements.

        Args:
            obj: the json object to decode
            element_type: a class object which is the conjure type of
                the elements in this list.
        Returns:
            A python list where the elements are instances of type
                element_type.
        """
        if not isinstance(obj, list):
            raise Exception("expected a python list")

        return list(map(lambda x: cls.do_decode(x, element_type), obj))

    @classmethod
    def decode_optional(
        cls, obj: Optional[Any], object_type: Type[DecodableType]
    ) -> Optional[Any]:
        """Decodes json into an element, returning None if the provided object
        is None.

        Args:
            obj: the json object to decode
            object_type: a class object which is the conjure type of
                the object if present.
        Returns:
            The decoded obj or None if no obj is provided.
        """
        if obj is None:
            return None

        return cls.do_decode(obj, object_type)

    @classmethod
    def decode_primitive(cls, obj, object_type):
        def raise_mismatch():
            raise Exception(
                "Expected to find {} type but found {} instead".format(
                    object_type, type(obj)
                )
            )

        if object_type is float:
            return float(obj)
        elif (
            object_type is str
            or object_type is BinaryType
            or isinstance(object_type, BinaryType)
        ):
            # Python 2/3 compatible way of checking string
            if not (
                isinstance(obj, str) or str(type(obj)) == "<type 'unicode'>"
            ):
                raise_mismatch()
        elif not isinstance(obj, object_type):
            raise_mismatch()

        return obj

    @classmethod
    def do_decode(cls, obj: Any, obj_type: Type[DecodableType]) -> Any:
        """Decodes json into the specified type

        Args:
            obj: the json object to decode
            obj_type: a class object which is the type we're decoding into.
        """

        type_origin = get_origin(obj_type)
        type_args = get_args(obj_type)

        if inspect.isclass(obj_type) and issubclass(obj_type, ConjureBeanType):
            return cls.decode_conjure_bean_type(obj, obj_type)

        elif inspect.isclass(obj_type) and issubclass(
            obj_type, ConjureUnionType
        ):
            return cls.decode_conjure_union_type(obj, obj_type)

        elif inspect.isclass(obj_type) and issubclass(
            obj_type, ConjureEnumType
        ):
            return cls.decode_conjure_enum_type(obj, obj_type)

        elif isinstance(obj_type, DictType):
            return cls.decode_dict(obj, obj_type.key_type, obj_type.value_type)

        elif isinstance(obj_type, ListType):
            return cls.decode_list(obj, obj_type.item_type)

        elif isinstance(obj_type, OptionalType):
            return cls.decode_optional(obj, obj_type.item_type)

        elif type_origin is OptionalTypeWrapper:
            return cls.decode_optional(obj, type_args[0])

        elif type_origin is dict:
            (key_type, value_type) = type_args
            return cls.decode_dict(obj, key_type, value_type)

        elif type_origin is list:
            return cls.decode_list(obj, type_args[0])

        return cls.decode_primitive(obj, obj_type)

    def decode(self, obj: Any, obj_type: Type[DecodableType]) -> Any:
        return self.do_decode(obj, obj_type)

    def read_from_string(
        self, string_value: str, obj_type: Type[DecodableType]
    ) -> Any:
        deserialized = json.loads(string_value)
        return self.decode(deserialized, obj_type)



# (c) Copyright 2018 Palantir Technologies Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#!/usr/bin/env python
from setuptools import find_packages, setup, Command
from os import path, makedirs, system
import subprocess
import sys

try:
    gitversion = (
        subprocess.check_output(
            "git describe --tags --always --first-parent".split()
        )
        .decode()
        .strip()
        .replace("-", "_")
    )
    open("conjure_python_client/_version.py", "w").write(
        '__version__ = "{}"\n'.format(gitversion)
    )
    if not path.exists("build"):
        makedirs("build")
except subprocess.CalledProcessError:
    print("outside git repo, not generating new version string")
exec(open("conjure_python_client/_version.py").read())


class FormatCommand(Command):
    """Enables setup.py format."""

    description = "Reformat python files using 'black'"
    user_options = [
        ("check", "c", "Don't write the files back, just return the status")
    ]

    def initialize_options(self):
        self.check = False

    def finalize_options(self):
        if self.check != False:
            self.check = True
        pass

    def run(self):
        try:
            if self.check:
                code = self.blackCheck()
            else:
                code = self.black()
            if code == 0:
                sys.exit(0)
            else:
                sys.exit(1)
        except OSError:
            pass

    def black(self):
        return system("black --line-length 79 .")

    def blackCheck(self):
        return system("black --check --quiet --line-length 79 .")


setup(
    name="conjure-python-client",
    version=__version__,
    description="Conjure Python Library",
    # The project's main homepage.
    url="https://github.com/palantir/conjure-python-client",
    author="Palantir Technologies, Inc.",
    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=find_packages(exclude=["test*", "integration*"]),
    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=["requests"],
    tests_require=["pytest", "pyyaml"],
    python_requires=">=3.8",
    cmdclass={"format": FormatCommand},
    package_data={
        "conjure_python_client": ["py.typed"],
    },
)
